@@ -12,14 +12,13 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from requests.adapters import HTTPAdapter
-from typing import TypeVar, Type, List, Optional, Dict
+from requests.adapters import HTTPAdapter, Response, CaseInsensitiveDict
+from typing import TypeVar, Type, List, Optional, Dict, Any, Union
 from requests.exceptions import HTTPError
 from requests.packages.urllib3.poolmanager import PoolManager
 from requests.packages.urllib3.util.ssl_ import create_urllib3_context
 from requests.packages.urllib3.util import Retry
 from .configuration import ServiceConfiguration
-from future.utils import raise_from
 
 import binascii
 import os
@@ -49,27 +48,30 @@ CIPHERS = (
 )
 
 
-TRACE_ID_HEADER = 'X-B3-TraceId'  # type: str
+TRACE_ID_HEADER: str = "X-B3-TraceId"
 TRACE_ID_RANDOM_BYTES = 8
 
 
-def fresh_trace_id():
-    # type: () -> bytes
+def fresh_trace_id() -> bytes:
     # returns a bytestring which is a valid zipkin trace id
     return binascii.hexlify(os.urandom(TRACE_ID_RANDOM_BYTES))
 
 
 class Service(object):
-    _requests_session = None  # type: requests.Session
-    _uris = None  # type: List[str]
-    _connect_timeout = None  # type: float
-    _read_timeout = None  # type: float
-    _verify = None  # type: str
+    _requests_session: requests.Session
+    _uris: List[str]
+    _connect_timeout: float
+    _read_timeout: float
+    _verify: str
 
     def __init__(
-        self, requests_session, uris, _connect_timeout, _read_timeout, _verify
-    ):
-        # type: (requests.Session, List[str], float, float, str) -> None
+        self,
+        requests_session: requests.Session,
+        uris: List[str],
+        _connect_timeout: float,
+        _read_timeout: float,
+        _verify: str,
+    ) -> None:
         self._requests_session = requests_session
         self._uris = uris
         self._connect_timeout = _connect_timeout
@@ -77,13 +79,11 @@ class Service(object):
         self._verify = _verify
 
     @property
-    def _uri(self):
-        # type: () -> str
+    def _uri(self) -> str:
         """returns a random uri"""
         return random.SystemRandom().choice(self._uris)
 
-    def _request(self, *args, **kwargs):
-        # type (Any) -> Response
+    def _request(self, *args, **kwargs) -> Response:
         """Make requests using configured :class:`requests.Session`.
         Any error details will be extracted to an :class:`HTTPError`
         which will contain relevant error details when printed."""
@@ -93,27 +93,25 @@ class Service(object):
             _response.raise_for_status()
         except HTTPError as e:
             if e.response is not None:
-                raise_from(ConjureHTTPError(e), e)
+                raise ConjureHTTPError(e) from e
             raise e
         return _response
 
-    def __repr__(self):
-        return '{}(requests_session={}, uris={})'.format(
-            self.__class__.__name__,
-            "requests.Session(...)",
-            repr(self._uris)
+    def __repr__(self) -> str:
+        return "{}(requests_session={}, uris={})".format(
+            self.__class__.__name__, "requests.Session(...)", repr(self._uris)
         )
 
-    def _amend_request_kwargs(self, kwargs):
-        for param_kind in ['headers', 'params']:
+    def _amend_request_kwargs(self, kwargs) -> None:
+        for param_kind in ["headers", "params"]:
             if param_kind in kwargs:
                 kwargs[param_kind] = _clean_params(kwargs[param_kind])
-        kwargs['timeout'] = (self._connect_timeout, self._read_timeout)
-        kwargs['verify'] = self._verify
+        kwargs["timeout"] = (self._connect_timeout, self._read_timeout)
+        kwargs["verify"] = self._verify
         _add_trace_id(kwargs)
 
 
-def _clean_params(params):
+def _clean_params(params: Dict[str, Any]) -> Dict[str, Any]:
     cleaned_params = {}
     for key, value in params.items():
         if value is None:
@@ -122,40 +120,45 @@ def _clean_params(params):
     return cleaned_params
 
 
-def _clean_param_value(value):
+def _clean_param_value(value: Any) -> Any:
+    if isinstance(value, list):
+        return [_clean_param_value(element) for element in value]
     if isinstance(value, bool):
         return str(value).lower()
     return str(value)
 
 
-def _add_trace_id(kwargs):
-    # type: (Dict) -> None
+def _add_trace_id(kwargs: Dict[str, Any]) -> None:
     # Adds the trace ID to the arguments
-    if 'headers' not in kwargs:
-        kwargs['headers'] = {}
-    kwargs['headers'][TRACE_ID_HEADER] = fresh_trace_id()
+    if "headers" not in kwargs:
+        kwargs["headers"] = {}
+    kwargs["headers"][TRACE_ID_HEADER] = fresh_trace_id()
 
 
 class RetryWithJitter(Retry):
-    '''
+    """
     Extends the standard urllib Retry in order to match conjure behaviour.
-    Retry times contain a uniform jitter. Additionally all http methods are
+    Retry times contain a uniform jitter. Additionally, all http methods are
     considered valid to retry.
-    '''
+    """
 
-    def _is_method_retryable(self, method):
+    @classmethod
+    def _is_method_retryable(cls, method):
         return True
 
     def get_backoff_time(self):
         jitter = random.random()
-        return jitter * super(RetryWithJitter, self).get_backoff_time()
+        return jitter * super().get_backoff_time()
 
 
 class RequestsClient(object):
-
     @classmethod
-    def create(cls, service_class, user_agent, service_config):
-        # type: (Type[T], str, ServiceConfiguration) -> T
+    def create(
+        cls,
+        service_class: Type[T],
+        user_agent: str,
+        service_config: ServiceConfiguration,
+    ) -> T:
         # setup retry to match java remoting
         # https://github.com/palantir/http-remoting/tree/3.12.0#quality-of-service-retry-failover-throttling
         retry = RetryWithJitter(
@@ -167,7 +170,7 @@ class RequestsClient(object):
         transport_adapter = TransportAdapter(max_retries=retry)
         # create a session, for shared connection polling, user agent, etc
         session = requests.Session()
-        session.headers = {"User-Agent": user_agent}
+        session.headers = CaseInsensitiveDict({"User-Agent": user_agent})
         if service_config.security is not None:
             verify = service_config.security.trust_store_path
         else:
@@ -187,7 +190,7 @@ class TransportAdapter(HTTPAdapter):
     """Transport adapter that allows customising ssl things"""
 
     def init_poolmanager(
-            self, connections, maxsize, block=False, **pool_kwargs
+        self, connections, maxsize, block=False, **pool_kwargs
     ):
         self._pool_connections = connections
         self._pool_maxsize = maxsize
@@ -204,83 +207,72 @@ class TransportAdapter(HTTPAdapter):
 
 
 class ConjureHTTPError(HTTPError):
-    """A an HTTPError from a Conjure Service with ``SerializableError``
+    """An HTTPError from a Conjure Service with ``SerializableError``
     attributes extracted from the response."""
 
-    _cause = None               # type: Optional[HTTPError]
-    _error_code = None          # type: str
-    _error_name = None          # type: str
-    _error_instance_id = None   # type: str
-    _parameters = None          # type: Dict[str, str]
-    _trace_id = None            # type: str
+    _cause: Optional[HTTPError]
+    _error_code: str
+    _error_name: str
+    _error_instance_id: str
+    _parameters: Dict[str, str]
+    _trace_id: str
 
-    def __init__(self, http_error):
-        # type (HTTPError) -> None
+    def __init__(self, http_error: HTTPError) -> None:
         self._cause = http_error
-        self._trace_id = http_error.response.headers.get('X-B3-TraceId')
+        self._trace_id = http_error.response.headers.get("X-B3-TraceId")
         try:
             detail = http_error.response.json()
             self._error_code = detail.get("errorCode")
             self._error_name = detail.get("errorName")
             self._error_instance_id = detail.get("errorInstanceId")
             self._parameters = detail.get("parameters", dict())
-            message = "{}. ErrorCode: '{}'. ErrorName: '{}'. " \
-                "ErrorInstanceId: '{}'. TraceId: '{}'. Parameters: {}" \
-                .format(
+            message = (
+                "{}. ErrorCode: '{}'. ErrorName: '{}'. "
+                "ErrorInstanceId: '{}'. TraceId: '{}'. Parameters: {}".format(
                     http_error,
                     self._error_code,
                     self._error_name,
                     self._error_instance_id,
                     self._trace_id,
-                    self._parameters
+                    self._parameters,
                 )
+            )
         except ValueError:
-            message = "{}. TraceId: '{}'. Response: '{}'"\
-                .format(
-                    http_error,
-                    self._trace_id,
-                    http_error.response.text
-                )
+            message = "{}. TraceId: '{}'. Response: '{}'".format(
+                http_error, self._trace_id, http_error.response.text
+            )
         super(ConjureHTTPError, self).__init__(
-            message,
-            request=http_error.request,
-            response=http_error.response
+            message, request=http_error.request, response=http_error.response
         )
 
     @property
-    def cause(self):
-        # type: () -> Optional[HTTPError]
+    def cause(self) -> Optional[HTTPError]:
         """The wrapped ``HTTPError`` that was the direct cause of
         the ``ConjureHTTPError``.
         """
         return self._cause
 
     @property
-    def error_code(self):
-        # type: () -> str
+    def error_code(self) -> str:
         """A fixed code word identifying the type of error."""
         return self._error_code
 
     @property
-    def error_name(self):
-        # type: () -> str
+    def error_name(self) -> str:
         """A fixed name identifying the error."""
         return self._error_name
 
     @property
-    def error_instance_id(self):
-        # type: () -> str
+    def error_instance_id(self) -> str:
         """A unique identifier for this error instance."""
         return self._error_instance_id
 
     @property
-    def parameters(self):
-        # type: () -> Dict[str, str]
+    def parameters(self) -> Dict[str, str]:
         """A set of parameters that further explain the error."""
         return self._parameters
 
     @property
-    def trace_id(self):
-        # type: () -> str
+    def trace_id(self) -> str:
         """The X-B3-TraceId for the request."""
         return self._trace_id
